/* -*- C++ -*-
 *******************************************************************
 *
 *
 * BmpUtil.cpp - Implementation of the CBmpUtil class
 *
 *
 *******************************************************************
 *
 * Bitmap utilities class
 *
 *******************************************************************
 *
 * Created 2000-12-28 by Jim Mason <jmason@sirius.com>
 *
 * LoadBmpImage function created 1998-08-05 by Zafir Anjum; see
 *      http://codeguru.earthweb.com/bitmap/bitmap_from_bmp.shtml
 *
 *******************************************************************
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 *******************************************************************
 */


#include "stdafx.h"
#include "BmpUtil.h"

// LoadBmpImage function authored by Zafir Anjum, 1998-08-05
//     URL:  http://codeguru.earthweb.com/bitmap/bitmap_from_bmp.shtml
BOOL CBmpUtil::LoadBmpImage(LPCSTR szBmpFile, HBITMAP* phBitmap, HPALETTE *phPal)
{
    HANDLE hFile;
    if(!(hFile = CreateFile(szBmpFile, GENERIC_READ, 0, NULL,
                                OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)))
        return FALSE;

    BITMAPFILEHEADER bmfHeader;

    // Read file header
    DWORD dwSize = 0;
    ReadFile(hFile, (LPSTR)&bmfHeader, sizeof(bmfHeader), &dwSize, NULL);
    if(dwSize != sizeof(bmfHeader)) {
        CloseHandle(hFile);
        return FALSE;
    }

    // File type should be 'BM'
    if(bmfHeader.bfType != ((WORD) ('M' << 8) | 'B')) {
        CloseHandle(hFile);
        return FALSE;
    }

    // If we're just testing for a valid BMP file, we're done
    if(phBitmap == NULL) {
        CloseHandle(hFile);
        return TRUE;
    }

    // Get length of the remainder of the file and allocate memory
    DWORD nPackedDIBLen = GetFileSize(hFile, NULL) - sizeof(BITMAPFILEHEADER);
    HGLOBAL hDIB = GlobalAlloc(GMEM_FIXED, nPackedDIBLen);
    if (hDIB == 0) {
        CloseHandle(hFile);
        return FALSE;
    }

    // Read the remainder of the bitmap file.
    ReadFile(hFile, (LPSTR)hDIB, nPackedDIBLen, &dwSize, NULL);
    if (dwSize != nPackedDIBLen) {
        GlobalFree(hDIB);
        CloseHandle(hFile);
        return FALSE;
    }


    BITMAPINFOHEADER &bmiHeader = *(LPBITMAPINFOHEADER)hDIB ;
    BITMAPINFO &bmInfo = *(LPBITMAPINFO)hDIB ;

    // If bmiHeader.biClrUsed is zero we have to infer the number
    // of colors from the number of bits used to specify it.
    int nColors = bmiHeader.biClrUsed ? bmiHeader.biClrUsed : 
    1 << bmiHeader.biBitCount;

    LPVOID lpDIBBits;
    if(bmInfo.bmiHeader.biBitCount > 8)
        lpDIBBits = (LPVOID)((LPDWORD)(bmInfo.bmiColors + bmInfo.bmiHeader.biClrUsed) + 
            ((bmInfo.bmiHeader.biCompression == BI_BITFIELDS) ? 3 : 0));
    else
        lpDIBBits = (LPVOID)(bmInfo.bmiColors + nColors);

    // Create the logical palette
    if(phPal != NULL) {
        // Create the palette
        if(nColors <= 256) {
            UINT nSize = sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * nColors);
            LOGPALETTE *pLP = (LOGPALETTE *) new BYTE[nSize];

            pLP->palVersion = 0x300;
            pLP->palNumEntries = nColors;

            for( int i=0; i < nColors; i++) {
                pLP->palPalEntry[i].peRed = bmInfo.bmiColors[i].rgbRed;
                pLP->palPalEntry[i].peGreen = bmInfo.bmiColors[i].rgbGreen;
                pLP->palPalEntry[i].peBlue = bmInfo.bmiColors[i].rgbBlue;
                pLP->palPalEntry[i].peFlags = 0;
            }

        *phPal = CreatePalette(pLP);

        delete[] pLP;
        }
    }

    HDC dc = GetDC(NULL);
    HPALETTE hOldPalette = NULL;
    if(phPal) {
        hOldPalette = SelectPalette(dc, *phPal, FALSE);
        RealizePalette(dc);
    }

    *phBitmap = CreateDIBitmap(dc,                  // handle to device context 
                                    &bmiHeader,     // pointer to bitmap size and format data 
                                    CBM_INIT,       // initialization flag 
                                    lpDIBBits,      // pointer to initialization data 
                                    &bmInfo,        // pointer to bitmap color-format data 
                                    DIB_RGB_COLORS);    // color-data usage 
    if(hOldPalette)
        SelectPalette(dc, hOldPalette, FALSE);

    ReleaseDC(NULL, dc);

    GlobalFree(hDIB);

    CloseHandle(hFile);

    return TRUE;
}

void CBmpUtil::FastDrawBitmap(HDC hDC, int xpos, int ypos,
                                    int nWidth, int nHeight,
                                    HDC hSrcDC, int xOrigin, int yOrigin,
                                    HDC hMaskDC, int xMask, int yMask)
{
    COLORREF crOldBack = SetBkColor(hDC, RGB(255,255,255));
    COLORREF crOldText = SetTextColor(hDC, RGB(0,0,0));

    // Blit using transparency mask
    BitBlt(hDC, xpos, ypos, nWidth, nHeight, hSrcDC, xOrigin, yOrigin, SRCINVERT);
    BitBlt(hDC, xpos, ypos, nWidth, nHeight, hMaskDC, xMask, yMask, SRCAND);
    BitBlt(hDC, xpos, ypos, nWidth, nHeight, hSrcDC, xOrigin, yOrigin, SRCINVERT);

    // Restore settings
    SetBkColor(hDC, crOldBack);
    SetTextColor(hDC, crOldText);
}

void CBmpUtil::DrawBitmap(HDC hDC, HBITMAP hBitmap, int xpos, int ypos,
                                int nWidth, int nHeight,
                                int xOrigin, int yOrigin)
{
    // Create DCs for the image and the mask
    HDC dcImage, dcMask;
    dcImage = CreateCompatibleDC(hDC);
    dcMask = CreateCompatibleDC(hDC);

    // Setup the image DC with the supplied bitmap
    HBITMAP hOldBmpImage = (HBITMAP)SelectObject(dcImage, hBitmap);

    // Create the monochrome mask bitmap and setup its DC
    HBITMAP bmpMask = CreateBitmap(nWidth, nHeight, 1, 1, NULL);
    HBITMAP hOldBmpMask = (HBITMAP)SelectObject(dcMask, bmpMask);

    // Build mask based on transparent colour
    SetBkColor(dcImage, GetPixel(dcImage, xOrigin, yOrigin));
    BitBlt(dcMask, 0, 0, nWidth, nHeight, dcImage, xOrigin, yOrigin, SRCCOPY);

    // Transparently render the bitmap using the mask
    FastDrawBitmap(hDC, xpos, ypos, nWidth, nHeight, dcImage,
                    xOrigin, yOrigin, dcMask);

    // Restore the DCs
    SelectObject(dcMask, hOldBmpMask);
    DeleteObject(bmpMask);
    DeleteDC(dcMask);
    SelectObject(dcImage, hOldBmpImage);
    DeleteDC(dcImage);
}

