/* -*- C++ -*-
 *******************************************************************
 *
 *
 * registry.cpp - Implementation of the Registry class
 *
 *
 *******************************************************************
 *
 * Registry access class
 *
 *******************************************************************
 *
 * Created 2000-12-28 by Jim Mason <jmason@sirius.com>
 *
 * Copyright (C) 2001 Jim Mason.
 *
 *******************************************************************
 *
 * This file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this file; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *******************************************************************
 */


#include "stdafx.h"
#include "registry.h"

Registry::Registry() {
}

void Registry::SetRegistryKey(LPCTSTR szAppName, LPCTSTR szRegistryKey) {
    ASSERT(m_szRegistryKey.IsEmpty());
    ASSERT(szAppName != NULL);
    ASSERT(szRegistryKey != NULL);

    m_szRegistryKey = szRegistryKey;
    m_szProfileName = szAppName;
}

HKEY Registry::GetAppRegistryKey() {
    ASSERT(!m_szRegistryKey.IsEmpty());
    ASSERT(!m_szProfileName.IsEmpty());

    HKEY hAppKey = NULL;
    HKEY hSoftKey = NULL;
    HKEY hCompanyKey = NULL;
    if (RegOpenKeyEx(HKEY_CURRENT_USER, _T("software"), 0, KEY_WRITE|KEY_READ,
                       &hSoftKey) == ERROR_SUCCESS) {
        DWORD dw;
        if (RegCreateKeyEx(hSoftKey, m_szRegistryKey, 0, REG_NONE,
                      REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, NULL,
                      &hCompanyKey, &dw) == ERROR_SUCCESS) {
            RegCreateKeyEx(hCompanyKey, m_szProfileName, 0, REG_NONE,
                REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, NULL,
                &hAppKey, &dw);
        }
    }
    if (hSoftKey != NULL)
        RegCloseKey(hSoftKey);
    if (hCompanyKey != NULL)
        RegCloseKey(hCompanyKey);

    return hAppKey;
}

HKEY Registry::GetSectionKey(LPCTSTR lpszSection) {
    ASSERT(lpszSection != NULL);

    HKEY hSectionKey = NULL;
    HKEY hAppKey = GetAppRegistryKey();
    if (hAppKey == NULL)
        return NULL;

    DWORD dw;
    RegCreateKeyEx(hAppKey, lpszSection, 0, REG_NONE,
        REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, NULL,
        &hSectionKey, &dw);
    RegCloseKey(hAppKey);
    return hSectionKey;
}

UINT Registry::GetProfileInt(LPCTSTR szSection, LPCTSTR szEntry, int nDefault) {
    ASSERT(szSection != NULL);
    ASSERT(szEntry != NULL);
    ASSERT(!m_szRegistryKey.IsEmpty());

    HKEY hSecKey = GetSectionKey(szSection);
    if (hSecKey == NULL)
        return nDefault;
    DWORD dwValue;
    DWORD dwType;
    DWORD dwCount = sizeof(DWORD);
    LONG lResult = RegQueryValueEx(hSecKey, (LPTSTR)szEntry, NULL, &dwType,
        (LPBYTE)&dwValue, &dwCount);
    RegCloseKey(hSecKey);
    if (lResult == ERROR_SUCCESS) {
        ASSERT(dwType == REG_DWORD);
        ASSERT(dwCount == sizeof(dwValue));
        return (UINT)dwValue;
    }
    return nDefault;
}

CStdString Registry::GetProfileString(LPCTSTR szSection, LPCTSTR szEntry,
                                        LPCTSTR szDefault) {
    ASSERT(szSection != NULL);
    ASSERT(szEntry != NULL);
    ASSERT(!m_szRegistryKey.IsEmpty());

    HKEY hSecKey = GetSectionKey(szSection);
    if (hSecKey == NULL)
        return szDefault;
    CStdString szValue;
    DWORD dwType, dwCount;
    LONG lResult = RegQueryValueEx(hSecKey, (LPTSTR)szEntry, NULL, &dwType,
        NULL, &dwCount);
    if (lResult == ERROR_SUCCESS) {
        ASSERT(dwType == REG_SZ);
        lResult = RegQueryValueEx(hSecKey, (LPTSTR)szEntry, NULL, &dwType,
            (LPBYTE)szValue.GetBuffer(dwCount/sizeof(TCHAR)), &dwCount);
        szValue.ReleaseBuffer();
    }
    RegCloseKey(hSecKey);
    if (lResult == ERROR_SUCCESS) {
        ASSERT(dwType == REG_SZ);
        return szValue;
    }
    return szDefault;
}

BOOL Registry::WriteProfileInt(LPCTSTR szSection, LPCTSTR szEntry,
                                 int nValue) {
    ASSERT(szSection != NULL);
    ASSERT(szEntry != NULL);
    ASSERT(!m_szRegistryKey.IsEmpty());

    HKEY hSecKey = GetSectionKey(szSection);
    if (hSecKey == NULL)
        return FALSE;
    LONG lResult = RegSetValueEx(hSecKey, szEntry, NULL, REG_DWORD,
        (LPBYTE)&nValue, sizeof(nValue));
    RegCloseKey(hSecKey);
    return lResult == ERROR_SUCCESS;
}

BOOL Registry::WriteProfileString(LPCTSTR szSection, LPCTSTR szEntry,
                                    LPCTSTR szValue) {
    ASSERT(szSection != NULL);
    ASSERT(!m_szRegistryKey.IsEmpty());

    LONG lResult;
    if (szEntry == NULL) {
        HKEY hAppKey = GetAppRegistryKey();
        if (hAppKey == NULL)
            return FALSE;
        lResult = ::RegDeleteKey(hAppKey, szSection);
        RegCloseKey(hAppKey);
    } else if (szValue == NULL) {
        HKEY hSecKey = GetSectionKey(szSection);
        if (hSecKey == NULL)
            return FALSE;
        // necessary to cast away const below
        lResult = ::RegDeleteValue(hSecKey, (LPTSTR)szEntry);
        RegCloseKey(hSecKey);
    } else {
        HKEY hSecKey = GetSectionKey(szSection);
        if (hSecKey == NULL)
            return FALSE;
        lResult = RegSetValueEx(hSecKey, szEntry, NULL, REG_SZ,
            (LPBYTE)szValue, (lstrlen(szValue)+1)*sizeof(TCHAR));
        RegCloseKey(hSecKey);
    }
    return lResult == ERROR_SUCCESS;
}

